// Copyleft (Cl) 1999 by Malekith / The Final Crash of Corp. System
// <malekith@topnet.pl> / <http://malekith.topnet.pl>
// This is part of gallery image viewer. See file copying-left for details.
// $Id: browser.h,v 2.8 1999/12/11 21:54:13 malekith Exp malekith $

#ifndef _Browser_H
#define _Browser_H

#ifndef _Image_H
#	include "image.h"
#endif

#ifndef __error_t_defined
typedef int error_t;
#define __error_t_defined
#endif
struct argp_state;
struct obstack;

class Rect {
public:
	int x1, y1, x2, y2;
	Rect(int xx1, int yy1, int xx2, int yy2) :
		x1(xx1), y1(yy1), x2(xx2), y2(yy2) { }
	Rect(int w=0, int h=0) : x1(0), y1(0), x2(w), y2(h) { }
	int w() { return x2-x1; }
	int h() { return y2-y1; }
	bool is_in(const Rect & r) {
		return (x1 >= r.x1 && y1 >= r.y1 && 
			x2 <= r.x2 && y2 <= r.y2);
	}
	void move(int dx, int dy) {
		x1 += dx; y1 += dy;
		x2 += dx; y2 += dy;
	}
	bool operator == (const Rect & r) {
		return (x1 == r.x1 && y1 == r.y1 && 
			x2 == r.x2 && y2 == r.y2);
	}
	friend inline Rect operator + (const Rect & l, const Rect & r)
	{
		return Rect(l.x1 <? r.x1, l.y1 <? r.y1,
			    l.x2 >? r.x2, l.y2 >? r.y2);
	}
};

class Browser {
	// config-state
	q_type quantize;
	d_type dither;
	bool fit_screen;
	bool quiet, verbose, debug;
	bool screen_messages, show_message_screen;
	int fade_frames;
	int scale;		// %
	int scale_over_100;	// when scale is > 100
	int scale_below_100;	// go figure
	// misc-state
	int last_action;
	Rect dirty; // used to update as less as possible
	Bitmap orig;
	Bitmap on_screen;
	const char * orig_file_name;
	byte current_palette[256*3];	
	byte screen_palette[256*3];
	char ** file_names;
	int nfiles;
	int screen_w, screen_h;
	int white;
	// data set by set_buffers
	int w, h, fs_w, fs_h;
	int type,bpp;
	byte *data;
	// data needed on startup and finish
	byte *font_buffer;
	int video_mode, scroll_step;
	int xres, yres, rbpp; // requested
	bool gfx_is_dead;
	// this are for comunication between selector() and read()
	char * name; // new'ed, /tmp/dsfds if gzip etc
	bool file_is_temp;
	// functions
	void set_version();
	void event_loop();
	void scale_to(int ns);
	void toggle_fit_screen();
	void show_and_recover();
	void handle_arrows();
	void update_screen(bool update_palette = false);
	void set_palette();
	Bitmap * handle_dirty_rect(Bitmap * tmp);
	void kill_gfx();
	void init_gfx();
	Image * selector(const char * name, bool force_magic = false);
	void display_formats();
	// command line
	void handle_config(int argc, char ** argv);
	static error_t parse(int key, char * arg, argp_state * s);
	void args_finish(argp_state * s, obstack * files);
	bool set_video();
	void message(const char * fmt, ... );
	void read();
	void fade(const byte * to);
	// progress bar
	int act_progress; // in %
	int range_l, range_r;
	int subrange_l, subrange_r;
	void draw_progress(int perc);
	
	void buffer_stuff();
	void do_quantize();
	void strip_alpha();
	void bs_8();
	void bs_15();
	void bs_16();
	void bs_24();
	void bs_32();
	int rgbcolor(int r, int g, int b);
public:
	void run(int argc, char ** argv);
	void sub_doing(int from, int to)
	{
		subrange_l = (range_r-range_l) * from / 100 + range_l;
		subrange_r = (range_r-range_l) * to / 100 + range_l;
	}
	void now_doing(int from, int to) 
	{ 
		range_l = from; 
		range_r = to; 
		sub_doing(0,100);
	}
	void finish_doing()
	{
		if (range_r == 100) 
			range_r = 0;
		now_doing(range_r, 100);
	}
	void finish_subdoing()
	{
		if (subrange_r == range_r) 
			subrange_r = range_l;
		subrange_l = subrange_r;
		subrange_r = range_r;
	}
	
	bool progress_bar(int a, int b); // a / b
	Browser();
	~Browser();
	
	void set_buffers(Image *me, int date_type, int w, int h, int perc);
};

#define msgf(args...) ({ if (!quiet) printf(args); })
#define msgvf(args...) ({ if (verbose) printf(args); })
#define msgdf(args...) ({ if (debug) printf(args); })

#endif // _Browser_H
