// Copyleft (Cl) 1999 by Malekith / The Final Crash of Corp. System
// <malekith@topnet.pl> / <http://malekith.topnet.pl>
// This is part of gallery image viewer. See file copying-left for details.
// $Id: cmdline.cc,v 2.10 1999/12/11 21:54:13 malekith Exp malekith $

#include "main.h"
#include "image.h"
#include "browser.h"

#include <unistd.h>
#include <obstack.h>

#include <vga.h>

#ifdef HAVE_ARGP_H
#	include <argp.h>
#else
#	include "sim_argp.cc"
#endif

// ---------------------------------------------------------------------------
extern int build_no;
extern char * build_time;
// ---------------------------------------------------------------------------
#define obstack_chunk_alloc xmalloc
#define obstack_chunk_free free
// ---------------------------------------------------------------------------
void * xmalloc(int n)
{
	void * p = malloc(n);
	if (p == null) {
		fprintf(stderr, "no memory (%d bytes) for obstack.\n",n);
		abort();
	}
	return p;
}
// ---------------------------------------------------------------------------
char * get_word(FILE * f)
{
	obstack o;
	obstack_init(&o);
	int c;
	while(isspace(c=fgetc(f))) { }
	for(;;) {
		if(c==EOF || isspace(c)) break;
		else
			obstack_1grow(&o,c);
		c = fgetc(f);
	}
	int len=obstack_object_size(&o);
	if(!len) return 0;
	obstack_1grow(&o,0);
	char * p = new char[len+1];
	strcpy(p,(char*)obstack_finish(&o));
	_obstack_free(&o,0);
	return p;
}
// ---------------------------------------------------------------------------
bool Browser::set_video()
{
	vga_modeinfo *m;
	int i;
	int max_mode = vga_lastmodenumber();
	if (yres == -1 && xres == -1)
		xres = DEFAULT_XRES;
	if (rbpp == -1)
		rbpp = DEFAULT_BPP;
	if (rbpp > 32 || rbpp < 8)
		throw xDead(_("wrong bpp number!"));
	int col = 1 << rbpp;
	for (i = 1; i <= max_mode; i++)
		if ((m = vga_getmodeinfo(i)) &&
		    vga_hasmode(i) &&
		    (xres == -1 || m->width == xres) && 
		    (yres == -1 || m->height == yres) && 
		    m->bytesperpixel >= 1 && m->colors == col)
			break;
	if (i > max_mode) {
		char *p;
		asprintf(&p,_("Mode x = %d, y = %d, bpp = %d not supported "
			      "on this video card and/or svgalib."),
			      xres,yres,rbpp);
		throw xDead(p);
	}
	msgvf("Setting mode %s.\n",vga_getmodename(video_mode));
	video_mode = i;
	// it's ok for my card ...
	if (scroll_step == -1)
		if (xres != -1)
			scroll_step = DEFAULT_STEP * (xres / 320) * 
				      (xres / 320) * (rbpp + 4) / 8;
		else
			scroll_step = DEFAULT_STEP * (yres / 200) * 
				      (yres / 200) * (rbpp + 4) / 8;
	return true;
}
// ---------------------------------------------------------------------------
void Browser::args_finish(argp_state * s, obstack * files)
{
	int n = obstack_object_size(files) / 4;
	if(n == 0) 
		argp_error(s,_("no files could been open for read"));
		// this means that some has been passed, but none access()'ed
	nfiles = n;
	char ** src = (char**) obstack_finish(files);
	file_names = new char* [n+1];
	char ** dst = file_names;
	while ( n-- > 0 ) 
		*dst++ = *src++;
	*dst = 0; // mark end
	_obstack_free(files,0);
	set_video();
}
// ---------------------------------------------------------------------------
static void dump_defines()
{
	printf(_("Compiled-in options:\n"));
	printf(
#	ifdef HAVE_LIBJPEG
	"HAVE_LIBJPEG "
#	endif
#	ifdef HAVE_LIBPNG
	"HAVE_LIBPNG "
#	endif
#	ifdef HAVE_LIBTIFF
	"HAVE_LIBTIFF "
#	endif
#	ifdef HAVE_ARGP_H
	"HAVE_ARGP_H "
#	endif
#	ifdef HAVE_LIBINTL_H
	"HAVE_LIBINTL_H "
#	endif
#	ifdef HAVE_LOCALE_H
	"HAVE_LOCALE_H "
#	endif
	"\n"
#	ifdef RUN_CONSOLECHARS
	"RUN_CONSOLECHARS (" CONSOLECHARS_FONT ") "
#	endif
#	ifdef RUN_SETFONT
	"RUN_SETFONT " 
#	endif
#	ifdef SETFONT_FONT
	"(" SETFONT_FONT ") "
#	endif
	"\n"
	"PREFIX=\"" PREFIX "\" DEFAULT_XRES=%d DEFAULT_BPP=%d\n"
	"DEFAULT_STEP=%d DEF_QUANTIZE=%d DEF_DITHER=%d\n",
	DEFAULT_XRES, DEFAULT_BPP,
	DEFAULT_STEP, (int)DEF_QUANTIZE, (int)DEF_DITHER);
	exit(0);
}
// ---------------------------------------------------------------------------
bool is_in(int n, int set[])
{
	for (int i=0; set[i]; i++)
		if (n == set[i])
			return true;
	return false;
}
// ---------------------------------------------------------------------------
error_t Browser::parse(int key, char * arg, argp_state * s) 
{
	Browser * t = (Browser *) s->input; // this
	static obstack files;
	switch(key) {
	case 'x':
		t->xres = atoi(arg);
		if (t->xres<=0)
			argp_error(s,_("xres can't be '%s'"),arg);
		break;
	case 'y':
		t->yres = atoi(arg);
		if (t->yres<=0)
			argp_error(s,_("yres can't be '%s'"),arg);
		break;
	case 'b':
		t->rbpp = atoi(arg);
		if (!is_in(t->rbpp, ((int[]){8,15,16,24,32,0}) ))
			argp_error(s,_("bpp can't be '%s'"),arg);
		break;
	case 'Q':
		t->quiet=true;
		break;
	case 's':
		if ( (t->scroll_step=atoi(arg)) == 0 )
			argp_error(s,_("step can't be '%s'"),arg);
		break;
	case 'B':
		t->fit_screen = false;
		break;
	case 'f':
		t->fade_frames = atoi(arg);
		break;
	case 'd':
		switch (tolower(*arg)) {
		case 'n':
		case '0': 
			t->dither = dNone; 
			break;
		case 'o':
		case '1': 
			t->dither = dOrdered; 
			break;
		case 'f':
		case '2': 
			t->dither = dFS; 
			break;
		default:
			argp_error(s, _("dither mode '%s' not supported"), arg);
			break;
		}
		break;
	case 'q':
		switch (tolower(*arg)) {
		case '3':
		case '0': 
			t->quantize = q332;
			break;
		case 'f':
		case '1': 
			t->quantize = qFast;
			break;
		case 's':
		case '2': 
			t->quantize = qSlow;
			break;
		default:
			argp_error(s, _("quantize mode '%s' not supported"), arg);
			break;
		}
	case 'D':
		t->debug = true;
	case 'v':
		t->verbose=true;
		break;
	case 'o':
		dump_defines();
		break;
	case ARGP_KEY_INIT:
		obstack_init(&files);
		break;
	case ARGP_KEY_FINI:
		t->args_finish(s,&files);
		break;
	case 'n':
		t->screen_messages = false;
		break;
	case 'p':
		t->show_message_screen = false;
		break;
	case 'r':
		t->display_formats();
		exit(0);
		break;
	case ARGP_KEY_ARG:
		if(access(arg,R_OK)) {
			if (!t->quiet)
				printf(_("warning: can't open '%s' for read\n"),
					arg);
		} else {
			char * p = new_str(arg);
			obstack_grow( &files, &p, sizeof(p) );
			// i used to get warnings if used obstack_ptr_grow
			// instead
		}
		break;
	case ARGP_KEY_NO_ARGS:
		argp_error(s,_("no images filenames passed"));
		break;
	default:
		return ARGP_ERR_UNKNOWN;
	}
	return 0;
}
// ---------------------------------------------------------------------------
argp_option opts[] = {
	{ 	"quiet", 	'Q', 0, 	0,	
		N_("Turns off messages.") },
	{	"verbose",	'v', 0,		0,
		N_("gives you a lot of nice messages.")},
	{	"debug",	'D', 0,		0,
		N_("enters debug mode. Some *nice* additional messages! I love "
		"when I've got a screen of text per image!") },
	{	"options",	'o', 0,		0,
		N_("Displays compiled-in options.") },
	{	"formats",	'r', 0,		0,
		N_("Displays supported formats.") },
	{	"no-texts",	'n', 0,		0,
		N_("turns off image file names and scale factor in gfx mode.") },
	{	"no-splash",	'p', 0,		0,
		N_("turns off that fancy color message screen. It will be "
		"black.") },
	{	"step",		's', "NUMBER",	0,	
		N_("Number of pixel per one step of scroling.") },
	{	"big-pic",	'B', 0,		0,	
		N_("Starts browser with fit-screen feature shut down.") },
	{	"dither",	'd', "MODE",	0,
		N_("sets dithering mode. MODE can be:\n"
		"``fs'' (``2'') to set Floyd-Steinberg error "
		"diffusion dithering (default)\n"
		"``order'' (``1'') to turn on simple ordered dithering\n"
		"``none'' (``0'') to disable dithering.\n" 
		"Only first chracter of MODE is significant.") },
	{	"quantize",	'q', "MODE",	0,
		N_("sets quantize mode. MODE can be:\n"
		"``slow'' (``2'') default. Best quality.\n"
		"``fast'' (``1'') one pass quantize.\n"
		"``332'' (``0'') Uses standard 332 palette. Fast & ugly.\n" 
		"Only first chracter of MODE is significant.") },
	{	"fade",		'f', "FRAMES",	0,
		N_("How many frames use to fade in / out effects. 0 disables. "
		"Good efects for about 5-10. No fade in RGB...") },
	{	"xres",		'x', "NUMBER",	0,
		N_("X resolution (screen WIDTH). It should be something "
		"about 320, 640, 800 and so on.") },
	{	"yres",		'y', "NUMBER",	0,
		N_("Y resolution (screen HEIGHT), e.g. 200, 480, 600.") },
	{	"bpp",		'b', "BITS-PER-PIXEL", 0,
		N_("this can be 8, 15, 16, 24, 32.") },
	{ 	0 }
};
// ---------------------------------------------------------------------------
char help[] = N_(
"Shows specified images.\x0b"
"Nice isn't it?\n"
"It's enough to tell this viewer that you want e.g. -x640 to make him guess,\n"
"that you also want -y480 and vice versa. MODE-X resolutions arn't supported.\n"
"If no mode mode can meet your request program will exit. No mode guessing.\n"
"Some combinations of quantize and dither arn't supported. In this case the\n"
"nearest combination available will be used. For details read the docs\n"
"(that mitological thing which doesn't exist)\n"
"\n"
"Enjoy!");
// ---------------------------------------------------------------------------
void Browser::handle_config(int argc, char ** argv)
{
	set_version();
	argp_program_bug_address="<" MAIL ">";
	
	obstack av;
	obstack_init(&av);
	// copy prog name
	{
	char * p = new_str(*argv++);
	obstack_grow(&av,&p,sizeof(p));
	}
	
	FILE * f = fopen("/etc/" NAME ".rc","r");
	if(f) {
		char * p;
		while( (p=get_word(f)) != 0 ) {
			obstack_grow(&av,&p,sizeof(p));
			argc++;
		}
		fclose(f);
	}
	
	char * name;
	asprintf(&name,"%s/." NAME ".rc",getenv("HOME"));
	if (name == null) abort();
	f = fopen(name,"r");
	free(name);
	if(f) {
		char * p;
		while( (p=get_word(f)) != 0 ) {
			obstack_grow(&av,&p,sizeof(p));
			argc++;
		}
		fclose(f);
	}
	
	while(*argv) {
		char * p = new_str(*argv++);
		obstack_grow(&av,&p,sizeof(p));
	} 
	obstack_grow(&av,&*argv,sizeof(*argv));  // null terminate
	
	argv = (char**)obstack_finish(&av);	
	
	xres = yres = rbpp = video_mode = -1;
	
	argp parser = { 
		opts, 
		Browser::parse, 
		N_("image-files..."), 
		_(help)
	};
	argp_parse(&parser,argc,argv,0,0,(void*)this);
	
	while(*argv) delete *argv++;
	_obstack_free(&av,0);
}
// ---------------------------------------------------------------------------
void Browser::set_version()
{
	char * p;
	asprintf(&p, _(
"%s %s\n"
"Build # %d, %s.\n"
"copyleft (Cl) 1999 %s <%s>\n"
"%s comes with NO WARRANTY, to the extent permitted by law.\n"
"You may redistribute copies of %s under the terms of the\n"
"GNU General Public License.\n"
"For more information about these matters, see the files named copying-left."
), 
NAME, VERSION, build_no, build_time, MAIL, HOME, NAME, NAME);
	assert(p);
	argp_program_version = p;
}
// ---------------------------------------------------------------------------
