/*         ______   ___    ___
 *        /\  _  \ /\_ \  /\_ \
 *        \ \ \L\ \\//\ \ \//\ \      __     __   _ __   ___
 *         \ \  __ \ \ \ \  \ \ \   /'__`\ /'_ `\/\`'__\/ __`\
 *          \ \ \/\ \ \_\ \_ \_\ \_/\  __//\ \L\ \ \ \//\ \L\ \
 *           \ \_\ \_\/\____\/\____\ \____\ \____ \ \_\\ \____/
 *            \/_/\/_/\/____/\/____/\/____/\/___L\ \/_/ \/___/
 *                                           /\____/
 *                                           \_/__/
 *      By Shawn Hargreaves,
 *      1 Salisbury Road,
 *      Market Drayton,
 *      Shropshire,
 *      England, TF9 1AJ.
 *
 *      LBM reader by Adrian Oboroc.
 *
 *	Copyright :
 *
 *  Allegro is swap-ware. You may use, modify, redistribute, and generally 
 *  hack it about in any way you like, but if you do you must send me 
 *  something in exchange. This could be a complimentary copy of a game, an 
 *  addition or improvement to Allegro, a bug report, some money (this is 
 *  particularly encouraged if you use Allegro in a commercial product), or 
 *  just a copy of your autoexec.bat if you don't have anything better. If 
 *  you redistribute parts of Allegro or make a game using it, it would be 
 *  nice if you mentioned me somewhere in the credits, but if you just want 
 *  to pinch a few routines that is OK too. I'll trust you not to rip me off.
 *
 */
 
/*
 * modified by Malekith
 * ilbm support has been striped out, mget[w|l] are now used instead of macros
 * $Id: iff-pbm.c,v 1.5 1999/12/11 21:54:13 malekith Exp malekith $
 */

#include "main-c.h"

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <limits.h>

#include "iff-pbm.h"

#define IFF_FORM     0x4D524F46	/* 'FORM' - IFF FORM structure  */
#define IFF_PBM      0x204D4250	/* 'PBM ' - new DP2e format     */
#define IFF_BMHD     0x44484D42	/* 'BMHD' - bitmap header       */
#define IFF_CMAP     0x50414D43	/* 'CMAP' - color map (palette) */
#define IFF_BODY     0x59444F42	/* 'BODY' - bitmap data         */

int progress_bar(int, int);

int igetw(FILE * f);
int igetl(FILE * f);
int mgetw(FILE * f)
{
	int b1, b2;
	if ((b1 = getc(f)) != EOF && (b2 = getc(f)) != EOF)
		return ((b1 << 8) | b2);
	return EOF;
}
int mgetl(FILE *f) {
   int b1, b2, b3, b4;

   if ((b1 = getc(f)) != EOF && (b2 = getc(f)) != EOF &&
       (b3 = getc(f)) != EOF && (b4 = getc(f)) != EOF )
	       return (((long)b1 << 24) | ((long)b2 << 16) |
		       ((long)b3 << 8) | (long)b4);
   return EOF;
}

int iff_pbm_get_header(char *filename, image_info *p)
{
	FILE *f = fopen(filename, "r");
	int id;
	if (!f)
		return 0;

	id = igetl(f);		/* read file header    */
	if (id != IFF_FORM) {	/* check for 'FORM' id */
		fclose(f);
		return 0;
	}
	igetl(f);		/* skip FORM length    */

	id = igetl(f);		/* read id             */

	/* check image type -- 'PBM ' */
	if (id != IFF_PBM) {
		fclose(f);
		return 0;
	}

	id = igetl(f);		/* read id               */
	if (id != IFF_BMHD) {	/* check for header      */
		fclose(f);
		return 0;
	}
	/* read header length    */
	if (mgetl(f) != 20) {	/* check, if it is right */
		fclose(f);
		return 0;
	}
	
	p->width = mgetw(f);		/* read screen width  */
	p->height = mgetw(f);		/* read screen height */
	
	mgetw(f);		/* skip initial x position  */
	mgetw(f);		/* skip initial y position  */

	p->bpp = getc(f);	/* get image depth   */
	fclose(f);
	
	if (p->bpp > 8)
		return 0;
	else
		return 1;
}

/* load_lbm:
 *  Loads IFF ILBM/PBM files with up to 8 bits per pixel, returning
 *  a bitmap structure and storing the palette data in the specified
 *  palette (this should be an array of at least 256 RGB structures).
 */
typedef unsigned char byte;
int iff_pbm_load(char *filename, int buf_size, byte *buf, byte *pal)
{
	FILE *f;
	int w, h, i, y, bpl;
	char ch, cmp_type, color_depth;
	unsigned char uc, check_flags, *line_buf;
	long id, len, l;

	f = fopen(filename, "r");
	if (!f)
		return 0;

	id = igetl(f);		/* read file header    */
	if (id != IFF_FORM) {	/* check for 'FORM' id */
		fclose(f);
		return 0;
	}
	igetl(f);		/* skip FORM length    */

	id = igetl(f);		/* read id             */

	/* check image type -- 'PBM ' */
	if (id != IFF_PBM) {
		fclose(f);
		return 0;
	}

	id = igetl(f);		/* read id               */
	if (id != IFF_BMHD) {	/* check for header      */
		fclose(f);
		return 0;
	}
	/* read header length    */
	if (mgetl(f) != 20) {	/* check, if it is right */
		fclose(f);
		return 0;
	}
	w = mgetw(f);		/* read screen width  */

	h = mgetw(f);		/* read screen height */
	
	if (buf_size < w * h) {
		fclose(f);
		return 0;
	}
	
	mgetw(f);		/* skip initial x position  */
	mgetw(f);		/* skip initial y position  */

	color_depth = getc(f);	/* get image depth   */
	if (color_depth > 8) {
		fclose(f);
		return 0;
	}
	getc(f);		/* skip masking type */

	cmp_type = getc(f);	/* get compression type */
	if ((cmp_type != 0) && (cmp_type != 1)) {
		fclose(f);
		return 0;
	}
	getc(f);		/* skip unused field        */
	mgetw(f);		/* skip transparent color   */
	getc(f);		/* skip x aspect ratio      */
	getc(f);		/* skip y aspect ratio      */
	mgetw(f);		/* skip default page width  */
	mgetw(f);		/* skip default page height */

	check_flags = 0;

	do {	/* We'll use cycle to skip possible junk      */
		/*  chunks: ANNO, CAMG, GRAB, DEST, TEXT etc. */
		id = igetl(f);

		switch (id) {
		case IFF_CMAP:
			l = mgetl(f);
			while (l-->0)
				*pal++ = getc(f);
			check_flags |= 1;	/* flag "palette read" */
			break;

		case IFF_BODY:
			mgetl(f);	/* skip BODY size */

			bpl = w;
			if (bpl & 1)	/* alignment            */
				bpl++;
			line_buf = malloc(bpl);
			if (!line_buf) {
				fclose(f);
				return 0;
			}
			for (y = 0; y < h; y++) {
				if (cmp_type) {
					i = 0;
					while (i < bpl) {
						uc = getc(f);
						if (uc < 128) {
							uc++;
							fread(buf, 1, uc, f);
							i += uc;
							buf += uc;
						} else if (uc > 128) {
							uc = 257 - uc;
							ch = getc(f);
							
							memset(buf, ch, uc);
							i += uc;
							buf += uc;
						}
						/* 128 (0x80) means NOP - no operation  */
					}
				} else	/* pure theoretical situation */
					fread(line_buf, 1, bpl, f);
				progress_bar(y, h);
			}
			check_flags |= 2;	/* flag "bitmap read" */
			break;

		default:	/* skip useless chunks  */
			len = mgetl(f);
			if (len & 1)
				len++;
			for (l = 0; l < (len >> 1); l++)
				mgetw(f);
		}

		/* Exit from loop if we are at the end of file, */
		/* or if we loaded both bitmap and palette      */

	} while ((check_flags != 3) && (!feof(f)));

	fclose(f);

	if (check_flags != 3)
		return 0;
	return 1;
}
