/* this module has been taken from */
/*
 * GTK See -- a image viewer based on GTK+
 * Copyright (C) 1998 Hotaru Lee <jkhotaru@mail.sti.com.cn> <hotaru@163.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/* 
 * and slighty modified by Malekith 
 * $Id: im_pcx.c,v 1.11 1999/12/12 10:09:41 malekith Exp malekith $
 */

#include "main-c.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "im_pcx.h"

int progress_bar(int,int);

gboolean	pcx_readline	(FILE *fd,
				 guchar* buffer,
				 gint bytes,
				 guint8 compression);

gboolean
pcx_get_header(gchar *filename, image_info *info)
{
	FILE *fd;
	pcx_header_struct pcx_header;
	
	fd = fopen (filename, "rb");
	if (fd == NULL)
	{
		return FALSE;
	}

	if (fread(&pcx_header, 128, 1, fd) == 0)
	{
		fclose(fd);
		return FALSE;
	}

	if(pcx_header.manufacturer != 10)
	{
		fclose(fd);
		return FALSE;
	}
	
	info->width = pcx_header.x2 - pcx_header.x1 + 1;
	info->height = pcx_header.y2 - pcx_header.y1 + 1;
	if (pcx_header.planes == 1 && pcx_header.bpp == 1)
	{
		info->bpp = 1;
	} else
	if (pcx_header.planes == 4 && pcx_header.bpp == 1)
	{
		info->bpp = 4;
	} else
	if (pcx_header.planes == 1 && pcx_header.bpp == 8)
	{
		info->bpp = 8;
	} else
	if (pcx_header.planes == 3 && pcx_header.bpp == 8)
	{
		info->bpp = 24;
	} else
	{
		fclose(fd);
		return FALSE;
	}
	
	fclose(fd);
	return TRUE;
}

gboolean
pcx_load(gchar *filename, int bsize, guchar * dbuf, guchar * palbuf)
{
	FILE *fd;
	pcx_header_struct pcx_header;
	gint width, height, bytes, x, y, c, t1, t2, pix, ptr;
	guchar *line, *line0, *line1, *line2, *line3;
	guchar cmap[768];
	
	fd = fopen (filename, "rb");
	if (fd == NULL)
	{
		return FALSE;
	}

	if (fread(&pcx_header, 128, 1, fd) == 0)
	{
		fclose(fd);
		return FALSE;
	}

	if(pcx_header.manufacturer != 10)
	{
		fclose(fd);
		return FALSE;
	}
	
	width = pcx_header.x2 - pcx_header.x1 + 1;
	height = pcx_header.y2 - pcx_header.y1 + 1;
	bytes = pcx_header.bytesperline;
	
	if (bsize < width * height)
	{
		fclose(fd);
		return FALSE;
	}
	if (pcx_header.planes == 1 && pcx_header.bpp == 1)
	{
		/* loading 1-bpp pcx */
		
		line = (guchar *) g_malloc (bytes * sizeof(guchar));
		for (y = 0; y < height; y++)
		{
			if (progress_bar(y, height)) break;
			if (!pcx_readline(fd, line, bytes, pcx_header.compression)) break;
			for (x = 0; x < width; x++) {
				if (line[x / 8] & (128 >> (x % 8)))
					*dbuf++ = 0x01;
				else
					*dbuf++ = 0x00;
			}
		}
		g_free(line);	
		*palbuf++ = 0x00;
		*palbuf++ = 0x00;
		*palbuf++ = 0x00;
		*palbuf++ = 0xff;
		*palbuf++ = 0xff;
		*palbuf++ = 0xff;
	} else
	if (pcx_header.planes == 4 && pcx_header.bpp == 1)
	{
		/* loading 4-bpp pcx */
		line0 = (guchar *) g_malloc (bytes * sizeof(guchar));
		line1 = (guchar *) g_malloc (bytes * sizeof(guchar));
		line2 = (guchar *) g_malloc (bytes * sizeof(guchar));
		line3 = (guchar *) g_malloc (bytes * sizeof(guchar));
		for (y = 0; y < height; y++)
		{
			if (progress_bar(y, height)) break;
			if (!pcx_readline(fd, line0, bytes, pcx_header.compression)) break;
			if (!pcx_readline(fd, line1, bytes, pcx_header.compression)) break;
			if (!pcx_readline(fd, line2, bytes, pcx_header.compression)) break;
			if (!pcx_readline(fd, line3, bytes, pcx_header.compression)) break;
			for (x = 0, ptr = 0; x < width; x++)
			{
				t1 = x / 8;
				t2 = (128 >> (x % 8));
				pix = (((line0[t1] & t2) == 0) ? 0 : 1) +
					(((line1[t1] & t2) == 0) ? 0 : 2) +
					(((line2[t1] & t2) == 0) ? 0 : 4) +
					(((line3[t1] & t2) == 0) ? 0 : 8);
				*dbuf++ = pix;
			}
		}
		g_free(line0);
		g_free(line1);
		g_free(line2);
		g_free(line3);
		memcpy(palbuf, pcx_header.colormap, sizeof(pcx_header.colormap));
	} else
	if (pcx_header.planes == 1 && pcx_header.bpp == 8)
	{
		/* loading 8-bpp pcx */
		fseek(fd, -768L, SEEK_END);
		if (fread(cmap, 768, 1, fd) == 0) return FALSE;
		fseek(fd, 128, SEEK_SET);
		if(bytes>width)bytes=width;
		for (y = 0; y < height; y++)
		{
			if (progress_bar(y, height)) break;
			if (!pcx_readline(fd, dbuf, bytes, pcx_header.compression)) break;
			dbuf += width;
		}
		memcpy(palbuf, cmap, sizeof(cmap));
	} else
	if (pcx_header.planes == 3 && pcx_header.bpp == 8)
	{
		if (bsize < width * height * 3) {
			fclose(fd);
			return FALSE;
		}			
		if(bytes>width*3)bytes=width*3;
		/* loading 24-bpp pcx(rgb) */
		line = (guchar *) g_malloc(bytes * 3 * sizeof(guchar));
		for (y = 0; y < height; y++)
		{
			if (progress_bar(y, height)) break;
			for (c= 0; c < 3; c++)
			{
				if (!pcx_readline(fd, line, bytes, pcx_header.compression)) break;
				for (x= 0; x < width; x++)
				{
					dbuf[x * 3 + c] = line[x];
				}
			}
			dbuf += 3 * width;
		}
		g_free(line);
	} else {
		fclose(fd);
		return FALSE;
	}

	fclose(fd);
	return TRUE;
}

gboolean
pcx_readline(FILE *fd, guchar *buffer, gint bytes, guint8 compression)
{
	guchar count = 0;
	gint value = 0;

	if (compression)
	{
		while (bytes--)
		{
			if (count == 0)
			{
				if ((value = fgetc(fd)) == EOF) return FALSE;
				if (value < 0xc0)
				{
					count= 1;
				} else
				{
					count= value - 0xc0;
					if ((value = fgetc(fd)) == EOF) return FALSE;
				}
			}
			count--;
			*(buffer++)= (guchar)value;
		}
	} else {
		if (fread(buffer, bytes, 1, fd) == 0) return FALSE;
	}
	return TRUE;
}
