/*
 * GTK See -- a image viewer based on GTK+
 * Copyright (C) 1998 Hotaru Lee <jkhotaru@mail.sti.com.cn> <hotaru@163.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/* 
 * slighty modified by Malekith
 * $Id: im_pnm.c,v 1.10 1999/11/09 20:46:37 malekith Exp malekith $
 */

#include "main-c.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "im_pnm.h"
int progress_bar(int,int);

gboolean
pnm_get_header(gchar *filename, image_info *info)
{
	FILE *file;
	guchar buffer[256];
	guchar type;

	file = fopen(filename, "r");
	if (file == NULL) return FALSE;
    
	if (!fgets(buffer, 255, file))
	{
		fclose(file);
		return FALSE;
	}
	
	if (buffer[0] != 'P' || (buffer[1] < '1' && buffer[1] > '6'))
	{
		fclose(file);
		return FALSE;
	}
	
	type = buffer[1];
	
	while (TRUE) {
		if (!fgets(buffer, 255, file))
		{
			fclose(file);
			return FALSE;
		}
		
		if (buffer[0]!='#') break;
	}
	
	if (sscanf(buffer, "%i %i", &info->width, &info->height) != 2)
	{
		fclose(file);
		return FALSE;
	}
	
	switch (type)
	{
	  case '1':
	  case '4':
		info -> bpp = 1;
		break;
	  case '2':
	  case '5':
		info -> bpp = 8;
		break;
	  case '3':
	  case '6':
		info -> bpp = 24;
		break;
	  default: /* never reach */
		fclose(file);
		return FALSE;
	}
	
	info -> slow = type >= '1' && type <= '3';
	
	return TRUE;
}
int pnm_get_int(FILE * f)
{
	int r = 0, c;
	while (isspace(c = fgetc(f))) {	}
	if (!isdigit(c))
		return EOF;
	while (isdigit(c)) {
		r *= 10;
		r += c - '0';
		c = fgetc(f);
	}
	return r;
}
gboolean
pnm_load(gchar *filename, int bufsize, guchar * imbuf, guchar * palbuf)
{
	FILE *file;
	guchar buffer[256];
	guchar type;
	gint width, height, max;
	gint i, j, c;
	guchar /* *dest, */  curbyte = 0;

	file = fopen(filename, "r");
	if (file == NULL) return FALSE;
    
	if (!fgets(buffer, 255, file))
	{
		fclose(file);
		return FALSE;
	}
	
	if (buffer[0] != 'P' || (buffer[1] < '1' && buffer[1] > '6'))
	{
		fclose(file);
		return FALSE;
	}
	
	type = buffer[1];
	
	while (TRUE) {
		if (!fgets(buffer, 255, file))
		{
			fclose(file);
			return FALSE;
		}
		
		if (buffer[0]!='#') break;
	}
	
	if (sscanf(buffer, "%i %i", &width, &height) != 2)
	{
		fclose(file);
		return FALSE;
	}
	
	if (!fgets(buffer, 255, file))
	{
		fclose(file);
		return FALSE;
	}
	
	if (sscanf(buffer, "%i", &max) != 1 || max < 1)
	{
		fclose(file);
		return FALSE;
	}
	// handle palette stuff
	if (bufsize < width * height) {
		fclose(file);
		return FALSE;
	}
	switch (type) {
	case '2':
	case '5':
		if (palbuf == NULL) {
			fclose(file);
			return FALSE;
		}
		for (i=0; i<256; i++) {
			*palbuf++ = i;
			*palbuf++ = i;
			*palbuf++ = i;
		}
		break;
	case '1':
	case '4':
		if (palbuf == NULL) {
			fclose(file);
			return FALSE;
		}
		*palbuf++=0;
		*palbuf++=0;
		*palbuf++=0;
		*palbuf++=0xff;
		*palbuf++=0xff;
		*palbuf++=0xff;
		break;
	case '3':
	case '6':
		if (bufsize < width * height * 3) {
			fclose(file);
			return FALSE;
		}
		break;
	}
	switch (type)
	{
	  case '2': /* ascii gray */
		//dest = g_malloc(sizeof(guchar) * width);
		for (i = 0; i < height; i++)
		{
			for (j = 0; j < width; j++)
			{
				//if (!fgets(buffer, 255, file)) { i = height; break; }
				//if (sscanf(buffer, "%i", &c) != 1) { i = height; break; }
				//dest[j] = (guchar)c;
				c = pnm_get_int(file);
				if (c==EOF) { i = height; break; }
				*imbuf++ = (guchar)c; 
			}
			if (i >= height) break;
			//if ((*func) (dest, width, 0, i, 1, 0, 0)) break;
			if (progress_bar(i,height)) break;
		}
		//g_free(dest);
		break;
	  case '3': /* ascii rgb */
		//dest = g_malloc(sizeof(guchar) * width * 3);
		for (i = 0; i< height; i++)
		{
			for (j = 0; j < width * 3; j++)
			{
				//if (!fgets(buffer, 255, file)) { i = height; break; }
				//if (sscanf(buffer, "%i", &c) != 1) { i = height; break; }
				//dest[j] = (guchar)c;
				c = pnm_get_int(file);
				if (c==EOF) { i = height; break; }
				*imbuf++ = (guchar)c;
			}
			if (i >= height) break;
			if (progress_bar(i,height)) break;
			//if ((*func) (dest, width, 0, i, 3, 0, 0)) break;
		}
		//g_free(dest);
		break;
	  case '5': /* raw gray */
		//dest = g_malloc(sizeof(guchar) * width);
		for (i = 0; i < height; i++)
		{
			if (!fread(imbuf, width, 1, file)) break;
			imbuf += width;
			if (progress_bar(i,height)) break;
			//if ((*func) (dest, width, 0, i, 1, 0, 0)) break;
		}
		//g_free(dest);
		break;
	  case '6': /* raw rgb */
		//dest = g_malloc(sizeof(guchar) * width * 3);
		for (i = 0; i< height; i++)
		{
			if (!fread(imbuf, width * 3, 1, file)) break;
			imbuf += width * 3;
			//if ((*func) (dest, width, 0, i, 3, 0, 0)) break;
			if (progress_bar(i,height)) break;
		}
		//g_free(dest);
		break;
	  case '1': /* ascii bitmap */
		//dest = g_malloc(sizeof(guchar) * width);
		for (i = 0; i < height; i++)
		{
			for (j = 0; j < width; j++)
			{
				if (!fgets(buffer, 255, file)) { i = height; break; }
				//dest[j] = (buffer[0] == '0') ? 0xff : 0x00;
				*imbuf++ = (buffer[0] == '0') ? 0x01 : 0x00;
			}
			if (i >= height) break;
			//if ((*func) (dest, width, 0, i, 1, 0, 0)) break;
			if (progress_bar(i,height)) break;
		}
		//g_free(dest);
		break;
	  case '4': /* raw bitmap */
		//dest = g_malloc(sizeof(guchar) * width);
		for (i = 0; i < height; i++)
		{
			for (j = 0; j < width; j++)
			{
				if (j % 8 == 0)
				{
					if ((c = fgetc(file)) == EOF) { i = height; break; }
					curbyte = (guchar)c;
				}
				//dest[j] = (curbyte & 0x80) ? 0x00 : 0xff;
				*imbuf++ = (curbyte & 0x80) ? 0x00 : 0x01;
				curbyte <<= 1;
			}
			if (i >= height) break;
			//if ((*func) (dest, width, 0, i, 1, 0, 0)) break;
			if (progress_bar(i,height)) break;
		}
		//g_free(dest);
		break;
	  default:
		fclose(file);
		return FALSE;
	}
	
	fclose(file);
	return TRUE;
}
