// Copyleft (Cl) 1999 by Malekith / The Final Crash of Corp. System
// <malekith@topnet.pl> / <http://malekith.topnet.pl>
// This is part of gallery image viewer. See file copying-left for details.
// Created on Aug 10 1999
// $Id: image.h,v 2.7 1999/12/11 21:54:13 malekith Exp malekith $

#ifndef _Image_H
#define _Image_H

extern int bytespp, bitspp;

class Bitmap {
	byte * buf; 
	int w, h;
	int wbpp; // w * bytespp
public:
	int x, y; // user-handled view point
	Bitmap() : buf(0), w(0), h(0), x(0), y(0)  { }
	~Bitmap();
	void free(); // safe tu call several times
	void alloc(int nw, int ny); // calls free() first
	void resize_from(const Bitmap & src);
	void resize_from(const Bitmap & src, int ww, int hh) 
	{
		alloc(ww,hh);
		resize_from(src);
	}
	void copy_from(const Bitmap & src);
	void put(); // at x, y
	void put_center(); // go figure
	void put(Bitmap & target);
	void put_center(Bitmap & target);
	int width() { return w; }
	int height() { return h; }
	int line_width() { return wbpp; }
	byte * data() { return buf; }
	void set_buffer(byte *nb, int ww, int hh) // buf must be newed
	{
		buf = nb; w = ww; h = hh;
		wbpp = w * bytespp;
	}
};

enum q_type {
	qSlow,
	qFast,
	q332	
};
enum d_type {
	dFS,
	dOrdered,
	dNone
};

class Image {
private:
	const char * patterns[10];
protected:
	virtual void do_read() = 0;
	// description of format
	const char * magic;
	const char * fmt_name; // e.g. PNG
	const char * fmt_cmt; // e.g. Portable Network Graphic, PNG's not gif
	const char * fmt_src; // source of loader, eg libpng by ...
public:
	Image(const char * p1, 
	      const char * p2 = null, 
	      const char * p3 = null, 
	      const char * p4 = null, 
	      const char * p5 = null, 
	      const char * p6 = null
	      );
	virtual ~Image() { clear_temp(); }
	
	bool file_is_temp;
	char *file_name;
	void read();
	
	// set by set_buffers
	byte *data;
	byte *palette;
	int data_size;
	int line_width;
	
	void clear_temp();
	
	// detection routines
	bool match_name(const char * fname);
	virtual bool match_magic(const byte * buffer);
	// description routines
	const char * get_fmt_name() { return fmt_name; }
	const char * get_fmt_cmt() { return fmt_cmt; }
	const char * get_fmt_src() { return fmt_src; }
};

enum {
	dtPalette,
	dtGray,
	dtRGB,
	dtRGBA,
	dtRGBA_ud // RGBA upside down (TIFF)
}; 
// sets data & palette buffers
void set_buffers(Image *me, int date_type, int w, int h, int perc); 

extern const byte black_palette[256*3];

#endif // _Image_H
