// Copyleft (Cl) 1999 by Malekith / The Final Crash of Corp. System
// <malekith@topnet.pl> / <http://malekith.topnet.pl>
// This is part of gallery image viewer. See file copying-left for details.
// Created on Aug 10 1999
// $Id: jpeg.cc,v 2.8 1999/12/11 21:54:13 malekith Exp malekith $

#include "main.h"

#ifdef HAVE_LIBJPEG
#include "jpeg.h"

// this people from IJG are relly funny, in one place in header they
// wrote: 'this would be a lot cleaner in C++' -- they were talking 
// about deriving implementation in C, it sound like : we like C++
// more, but we wrote this lib in C, for people using it ...
// but about they didn't think about extern "C" ...
//
// anyway: I want to greet them!
extern "C" {
#include <jpeglib.h>
}
#include <setjmp.h>

extern "C" int progress_bar(int,int);

// ---------------------------------------------------------------------------
JpegImage::JpegImage() : Image("*.jpeg", "*.jpg", "*.jpe"), 
	cinfo(new jpeg_decompress_struct) 
{
	magic = "\xff\xd8JFIF";
	fmt_name = "JPEG";
	fmt_cmt = "JPEG/JFIF";
	fmt_src = N_("libjpeg by Independent JPEG Group");
}
// ---------------------------------------------------------------------------
bool JpegImage::match_magic(const char *buffer)
{
	return memcmp("\xff\xd8",buffer,2) == 0 &&
	       memcmp("JFIF",buffer+6,4);
}
// ---------------------------------------------------------------------------
JpegImage::~JpegImage()
{
		delete cinfo;
}
// ---------------------------------------------------------------------------
void JpegImage::process()
{
	byte * samp[4];
	byte * dest = data;
	
	while( cinfo->output_scanline < cinfo->output_height ) {
		samp[0] = dest; dest += line_width;
		samp[1] = dest; dest += line_width;
		samp[2] = dest; dest += line_width;
		samp[3] = dest; dest += line_width;
		int n = jpeg_read_scanlines(cinfo, samp, 4);
		if(n<4) dest -= (4-n) * line_width;
		if (progress_bar(cinfo->output_scanline, cinfo->output_height)) 
			break;
	}
}
// ---------------------------------------------------------------------------
void JpegImage::jpeg_read() {
	if(jpeg_read_header(cinfo,true)) {
		jpeg_start_decompress(cinfo);
		
		if (cinfo->out_color_components == 1)
			set_buffers(this, dtGray, cinfo->image_width, 
				    cinfo->image_height, 30);
		else
			set_buffers(this, dtRGB, cinfo->image_width, 
				    cinfo->image_height, 50);
		
		process();
		jpeg_abort_decompress(cinfo);
	}		
	else throw xJpeg(_("can't read jpeg header"));
}
// ---------------------------------------------------------------------------
// error handling
struct my_jpeg_error_mgr : public jpeg_error_mgr {
	char msg[JMSG_LENGTH_MAX+20];
	jmp_buf jmp;
};
static void jpeg_err_handler(j_common_ptr cinfo)
{
	char buffer[JMSG_LENGTH_MAX];	
	(*cinfo->err->format_message) (cinfo, buffer);
	
	my_jpeg_error_mgr * em = (my_jpeg_error_mgr*) cinfo->err;
	sprintf(em->msg, _("jpeg error: %s"), buffer);
	longjmp(em->jmp,1);
}
// ---------------------------------------------------------------------------
void JpegImage::do_read()
{
	FILE * f = fopen(file_name,"r");
	if(!f) throw xJpeg(_("can't open file for reading"));
	try {
		my_jpeg_error_mgr jerr;
		cinfo->err = jpeg_std_error(&jerr);
		jerr.error_exit = jpeg_err_handler;
		
		if(setjmp(jerr.jmp)) 
			throw xJpeg(jerr.msg);
		
		jpeg_create_decompress(cinfo);
		jpeg_stdio_src(cinfo,f);
		
		jpeg_read();
		
		jpeg_destroy_decompress(cinfo);
	} catch(xcpt & e) {
		fclose(f);
		throw e;
	}
	fclose(f);
}
// ---------------------------------------------------------------------------

#endif // HAVE_LIBJPEG
