// Copyleft (Cl) 1999 by Malekith / The Final Crash of Corp. System
// <malekith@topnet.pl> / <http://malekith.topnet.pl>
// This is part of gallery image viewer. See file copying-left for details.
// Created on Jul 30 1999
// $Id: main.cc,v 2.16 1999/12/11 21:54:13 malekith Exp malekith $

#include "main.h"
#include "image.h"
#include "browser.h"

#include <sys/time.h>
#include <unistd.h>

#include <vga.h>
#include <vgagl.h>
#include <vgakeyboard.h>

// ---------------------------------------------------------------------------
// global functions
// ---------------------------------------------------------------------------
const int ups=1000;
int bytespp;
int bitspp;
extern "C" int getclock();
// ---------------------------------------------------------------------------
// Browser functions
// ---------------------------------------------------------------------------
Browser::Browser() : 
	quantize(DEF_QUANTIZE),
	dither(DEF_DITHER),
	fit_screen(true),
	quiet(false), verbose(false), debug(false), 
	screen_messages(true), show_message_screen(true), 
	fade_frames(7),
	scale(100), 
	scale_over_100(100), scale_below_100(10),
	
	last_action(-2*ups),
	orig_file_name(null),
	file_names(null),
	white(0xff),
	
	font_buffer(null),
	video_mode(-1), scroll_step(-1),
	gfx_is_dead(true),
	
	name(null),
	file_is_temp(false)
{
	memset(screen_palette, 0, sizeof(screen_palette));
}
// ---------------------------------------------------------------------------
Browser::~Browser()
{
	kill_gfx();
}
// ---------------------------------------------------------------------------
void Browser::init_gfx()
{
	msgf(	_("%s, Copyleft (Cl) 1999 Malekith\n"
		"Ptrs to author: <%s> / <%s>\n"
		"This program falls under GNU GPL v2.\n"),
		FULL_NAME, MAIL, HOME);
		
	// if PARANOID_SECURITY is set this is done as a first func call
#	ifndef PARANOID_SECURITY
	if(quiet) vga_disabledriverreport(); // shut up!
	if(vga_init()) throw xDead(_("can't vga_init()"));
#	endif
	
	set_video();
	int bufsize = vga_ext_set(VGA_EXT_FONT_SIZE,0);
	if(!bufsize) 
		throw xDead("can't read font buffer len");
	font_buffer = new byte[bufsize];
	
	if(vga_setmode(video_mode)) 
		throw xDead(_("can't vga_setmode()"));
	gfx_is_dead = false;
	vga_gettextfont(font_buffer);
	if(gl_setcontextvga(video_mode)) 
		throw xDead("can't gl_setcontextvga()");
	if(keyboard_init()) 
		throw xDead("can't kayboard_init()");
	if (rbpp == 8)
		gl_setpalette((void*)black_palette);
	gl_clearscreen(0);
	gl_setfont(8, 8, gl_font8x8);
	gl_setwritemode(FONT_COMPRESSED|WRITEMODE_MASKED);
	
	if (rbpp == 8) {
		gl_setfontcolors(0, white);
		gl_setpalettecolor(white, 0xff, 0xff, 0xff);
	} else
		gl_setfontcolors(0, gl_rgbcolor(0xff,0xff,0xff));
	
	screen_w = WIDTH;
	screen_h = HEIGHT;
	bytespp = BYTESPERPIXEL;
	bitspp = rbpp;
	dirty = Rect(screen_w, screen_h);
}
// ---------------------------------------------------------------------------
void Browser::kill_gfx()
{
	if (gfx_is_dead)
		return;
	keyboard_close();
	vga_puttextfont(font_buffer);
	vga_setmode(TEXT);
#	ifdef RUN_CONSOLECHARS
	// this is to set font ok on redhat & my strange card ...
	system("/usr/bin/consolechars -f " CONSOLECHARS_FONT);
#	elif defined RUN_SETFONT
	// this works on slak4 with my strange card
	system("/usr/bin/setfont"
#		ifdef SETFONT_FONT
		" " SETFONT_FONT
#		endif
	);
#	endif
	gfx_is_dead = true;
}
// ---------------------------------------------------------------------------
Bitmap *Browser::handle_dirty_rect(Bitmap * tmp)
{
	Rect update;
	Rect image(on_screen.width(), on_screen.height());
	Rect under_text(0, 0, screen_w, 9);
	bool small = image.is_in(Rect(screen_w, screen_h));
	// small means centred
	if (small)
		image.move((screen_w - image.w()) / 2,
			   (screen_h - image.h()) / 2);

	update = dirty + image;
	dirty = image;

	if (!under_text.is_in(update)) {
		gl_fillbox(under_text.x1, under_text.y1, under_text.x2,
			   under_text.y2, 0);
	}

	if (small) {
		tmp->alloc(update.w(), update.h());
		tmp->x = -update.x1;
		tmp->y = -update.y1;

		on_screen.x = image.x1 - update.x1;
		on_screen.y = image.y1 - update.y1;
		on_screen.put(*tmp);
		return tmp;
	} else {
		if (!update.is_in(image))
			gl_clearscreen(0);
		return &on_screen;
	}
}
// ---------------------------------------------------------------------------
void Browser::set_palette()
{
	if (bitspp != 8) {
		white = gl_rgbcolor(0xff,0xff,0xff);
		return;
	}
	byte * p = current_palette;
	p[0] = p[1] = p[2] = 0;
	fade(current_palette);
	// find brightest color
	int mc = -1;
	for(int i=0; i<256; i++) {
		int k = *p++;
		k += *p++; 
		k += *p++;
		if (k > mc) { 
			mc = k; 
			white = i; 
		}
	}
}
// ---------------------------------------------------------------------------
void Browser::fade(const byte * to) 
{
	if (fade_frames != 0) {
		byte cur_pal[256*3];
		const int fr = fade_frames;
		for(int i=0; i<fr; i++) {
			float nf = (float) (fr-i-1) / fr;
			float nt = 1 - nf;
			int k = sizeof(cur_pal);
			const byte * f = screen_palette;
			const byte * d = to;
			byte * p = cur_pal;
			while (k-->0)
				*p++ = (byte)(*f++ * nf + *d++ * nt);
			vga_waitretrace();
			gl_setpalette((void*)cur_pal);
		}
	} else
		gl_setpalette((void*)to);
	memcpy(screen_palette,to,sizeof(screen_palette));
}
// ---------------------------------------------------------------------------
void Browser::update_screen(bool update_pal)
{
	Bitmap tmp;
	// it can return tmp or on_screen
	Bitmap *put_this = handle_dirty_rect(&tmp);

	if (bitspp == 8 && update_pal)
		fade(black_palette);

	put_this->put();

	if (update_pal) {
		set_palette();
		gl_setfontcolors(0, white);
	}
}
// ---------------------------------------------------------------------------
int Browser::rgbcolor(int r, int g, int b)
{
	if (bitspp == 8) {
		int dmin = 0;
		byte *p = current_palette + 3;
		int m = -1;
		for (int i=1; i<256; i++) {
			int k=0;
			k += (*p - r) * (*p - r); p++;
			k += (*p - g) * (*p - g); p++;
			k += (*p - b) * (*p - b); p++;
			if (m == -1 || dmin > k) {
				m = i;
				dmin = k;
			}
		}
		return m;
	} else {
		int x = gl_rgbcolor(r,g,b);
		if (!x)
			x++;
		return x;
	}
}
// ---------------------------------------------------------------------------
void Browser::handle_arrows()
{
	if (on_screen.width() <= screen_w && on_screen.height() <= screen_h)
		return; // picture fits screen and shouldn't be moved
	int big_pic_x = on_screen.x, big_pic_y = on_screen.y;
	if (keyboard_keypressed(SCANCODE_CURSORBLOCKLEFT))
		big_pic_x -= scroll_step;
	if (keyboard_keypressed(SCANCODE_CURSORBLOCKRIGHT))
		big_pic_x += scroll_step;
	if (keyboard_keypressed(SCANCODE_CURSORBLOCKUP))
		big_pic_y -= scroll_step;
	if (keyboard_keypressed(SCANCODE_CURSORBLOCKDOWN))
		big_pic_y += scroll_step;
		
	big_pic_x <?= on_screen.width()  - WIDTH - 1;
	big_pic_y <?= on_screen.height() - HEIGHT - 1;
	big_pic_x >?= 0;
	big_pic_y >?= 0;
	
	if (on_screen.x != big_pic_x || on_screen.y != big_pic_y) {
		on_screen.x = big_pic_x;
		on_screen.y = big_pic_y;
		on_screen.put();
	}
}
// ---------------------------------------------------------------------------
void Browser::read()
{
	Image *loader = selector(orig_file_name);	
	loader->file_is_temp = file_is_temp;
	loader->file_name = name;
	loader->read();
	loader->clear_temp();
	buffer_stuff();
}
// ---------------------------------------------------------------------------
void Browser::show_and_recover()
{
	try {
		msgvf(_("loading %s...\n"), orig_file_name);
		on_screen.x = 0;
		on_screen.y = 0;

		read();

		scale = on_screen.width() * 100 / orig.width();
		update_screen(true);
		if (screen_messages)
			gl_printf(0, 0, "%s (%d%%)", orig_file_name, scale);
		msgvf(_("ok\n"));
	}
	catch (xRecover & e) {
		message(_("Error ``%s'' loading file ``%s''. "
			"Press space for next image."), e.msg, orig_file_name);
	}
}
// ---------------------------------------------------------------------------
void Browser::toggle_fit_screen()
{
	fit_screen = !fit_screen;
	last_action = getclock();
	if (fit_screen) {
		on_screen.resize_from(orig,fs_w,fs_h);
		scale = on_screen.width() * 100 / orig.width();
	} else {
		on_screen.copy_from(orig);
		scale = 100;
	}
	update_screen();	
}
// ---------------------------------------------------------------------------
void Browser::scale_to(int ns)
{
	if (ns <= 0) return;
	last_action = getclock();
	// remember relative pos
	scale = ns;
	if (on_screen.width() && on_screen.height()) {
		int px = on_screen.x * 10000 / on_screen.width();
		int py = on_screen.y * 10000 / on_screen.height();
	
		on_screen.resize_from(orig,w*scale/100,h*scale/100);
		on_screen.x = px * on_screen.width()  / 10000;
		on_screen.y = py * on_screen.height() / 10000;
	} else {
		on_screen.resize_from(orig,w*scale/100,h*scale/100);
		on_screen.x = on_screen.y = 0;
	}
	
	fit_screen = false;
	
	update_screen();
}
// ---------------------------------------------------------------------------
void Browser::event_loop() 
{
	int last=-1;
	for(int i=0;;) {
		if (i >= nfiles) break;
		if (i != last) {
			last_action=getclock();
			orig_file_name = file_names[i];
			show_and_recover();
			last=i;
		}
		usleep(10000);
		keyboard_update();
		if (keyboard_keypressed(SCANCODE_Q) ||
		    keyboard_keypressed(SCANCODE_ESCAPE)) break;
		handle_arrows();
		
		if (getclock()-last_action>ups/2) {
			if (keyboard_keypressed(SCANCODE_BACKSPACE) && i>0) 
				i--;
			if (keyboard_keypressed(SCANCODE_SPACE))
				i++;
				
			// scaling
			if (keyboard_keypressed(SCANCODE_ENTER)) 
				toggle_fit_screen();				
			if (keyboard_keypressed(SCANCODE_EQUAL) || 
			    keyboard_keypressed(SCANCODE_KEYPADPLUS))
			    	scale_to(scale + (scale < 100 ? 
							scale_below_100 :
							scale_over_100));
			if (keyboard_keypressed(SCANCODE_MINUS) || 
			    keyboard_keypressed(SCANCODE_KEYPADMINUS))
			    	scale_to(scale - (scale <= 100 ? 
							scale_below_100 :
							scale_over_100));
			int n = -1;
			for (int i=SCANCODE_1; i <= SCANCODE_0; i++)
				if (keyboard_keypressed(i))
					n = i - SCANCODE_1 + 1;
			if (n != -1) {
				int s;
				if (keyboard_keypressed(SCANCODE_LEFTSHIFT) ||
				    keyboard_keypressed(SCANCODE_RIGHTSHIFT))
				    	s = 10 * n;
				else
					s = 100 * n;
				scale_to(s);
			}				
		}
	}
}
// ---------------------------------------------------------------------------
void Browser::run(int argc, char ** argv)
{
#	ifdef HAVE_LOCALE_H
	setlocale(LC_ALL, "");
#	endif
#	ifdef HAVE_LIBINTL_H
	bindtextdomain(NAME, LOCALEDIR);
	textdomain(NAME);
#	endif
#	ifdef HAVE_LOCALE_H
	setlocale(LC_ALL, "");
#	endif
	handle_config(argc,argv);
	init_gfx();
	event_loop();
	kill_gfx();
	char ** p = file_names;
	while(*p) delete [] *p++;
	delete [] file_names;
	delete [] font_buffer;
}
// ---------------------------------------------------------------------------
bool Browser::progress_bar(int a, int b)
{
	if (a < 0 && b < 0) {
		a = -a;
		b = -b;
	}
	
	int np = (a >? 0) * 100 / (b >? 1);
	np <?= 100;
	np = (subrange_r - subrange_l) * np / 100 + subrange_l;
	if (np != act_progress)
		draw_progress(act_progress = np);
		
	keyboard_update();
	if (keyboard_keypressed(SCANCODE_Q) ||
	    keyboard_keypressed(SCANCODE_ESCAPE)) 
	    	return true;
	else
		return false;
}
// ---------------------------------------------------------------------------
void Browser::draw_progress(int perc)
{
	const int pbar_w = screen_w / 3;
	const int pbar_left = screen_w - 1 - pbar_w;
	const int pbar_right = pbar_left + pbar_w;
	const int pbar_y = screen_h - 1;
	int pbar_m = (perc * pbar_w / 100 + pbar_left) <? pbar_right;
	
	gl_hline(pbar_left, pbar_y, pbar_m, white);
	gl_hline(pbar_m, pbar_y, pbar_right, 0);
}
// ---------------------------------------------------------------------------
Browser * browser_ptr;
extern "C" int progress_bar(int a, int b)
{
	return (int) browser_ptr->progress_bar(a, b);
}
extern "C" void now_doing(int a, int b)
{
	browser_ptr->now_doing(a, b);
}
extern "C" void sub_doing(int a, int b)
{
	browser_ptr->sub_doing(a, b);
}
extern "C" void finish_doing()
{
	browser_ptr->finish_doing();
}
extern "C" void finish_subdoing()
{
	browser_ptr->finish_subdoing();
}
void set_buffers(Image *me, int data_type, int w, int h, int perc)
{
	browser_ptr->set_buffers(me,data_type,w,h,perc);
}
// ---------------------------------------------------------------------------
int main(int argc, char ** argv)
{
#	ifdef PARANOID_SECURITY
	vga_disabledriverreport();
	if (vga_init()) {
		fprintf(stderr,"can't vga_init()\n");
		return 1;
	}
#	endif
	try {
		{
			Browser browser;
			browser_ptr = &browser;
			browser.run(argc,argv);
		}
	} catch(xcpt & e) {
		fprintf(stderr,"%s: '%s'\n",e.name(),e.msg);
		abort();
	} catch(...) {
		fprintf(stderr,"unknown unhandled xcpt\n");
		abort();
	}
	return 0;
}
// ---------------------------------------------------------------------------
