// Copyleft (Cl) 1999 by Malekith / The Final Crash of Corp. System
// <malekith@topnet.pl> / <http://malekith.topnet.pl>
// This is part of gallery image viewer. See file copying-left for details.
// Created on Aug 10 1999
// $Id: png.cc,v 2.4 1999/12/11 21:54:13 malekith Exp malekith $

#include "main.h"

#ifdef HAVE_LIBPNG
#include "png.h"

#include <png.h>

extern "C" int progress_bar(int,int);
extern "C" int now_doing(int,int);

// ---------------------------------------------------------------------------
PngImage::PngImage() : Image("*.png") 
{ 
	magic = "\x89PNG\r\n\x1a\n"; 
	fmt_name = "PNG";
	fmt_cmt = "Portable Network Graphics or PNG's not GIF";
	fmt_src = N_("libpng by Guy Eric Schalnat, Andreas Dilger & Glenn Randers-Pehrson");
}
// ---------------------------------------------------------------------------
void PngImage::init()
{
	if ((fp = fopen(file_name, "rb")) == NULL)
		throw xPng(_("can't open file for reading"));

	png_ptr =
	    png_create_read_struct(PNG_LIBPNG_VER_STRING, null, null,
				   null);

	if (png_ptr == NULL) {
		fclose(fp);
		throw xPng("png_create_read_struct()");
	}

	info_ptr = png_create_info_struct(png_ptr);
	if (info_ptr == NULL) {
		fclose(fp);
		png_destroy_read_struct(&png_ptr, (png_infopp) NULL,
					(png_infopp) NULL);
		throw xPng("png_create_info_struct()");
	}
	if (setjmp(png_ptr->jmpbuf)) {
		png_destroy_read_struct(&png_ptr, &info_ptr,
					(png_infopp) NULL);
		fclose(fp);
		throw xPng(_("longjmp() from pnglib"));
	}

	png_init_io(png_ptr, fp);
	png_read_info(png_ptr, info_ptr);
}
// ---------------------------------------------------------------------------
void PngImage::set_filters()
{
}
// ---------------------------------------------------------------------------
void PngImage::read_file()
{
	assert((int) png_get_rowbytes(png_ptr, info_ptr) == line_width);
	
	byte *im = data;

	int image_height=png_ptr->height;
	png_ptr->num_rows = image_height; /* Make sure this is set correctly */
	
	int pass = png_set_interlace_handling(png_ptr);
	int m = pass * image_height, n = 0;
	
	for (int j = 0; j < pass; j++) {
		byte *p = im;
		for (int i = 0; i < image_height; i++) {
			png_read_row(png_ptr, p, NULL);
			p += line_width;
			progress_bar(n++,m);
		}
	}
}
// ---------------------------------------------------------------------------
void PngImage::free()
{
	png_read_end(png_ptr, info_ptr);
	png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp) NULL);
	fclose(fp);
}
// ---------------------------------------------------------------------------
void PngImage::do_read()
{
	png_uint_32 width, height;
	int bit_depth, color_type, interlace_type;

	init();

	png_get_IHDR(png_ptr, info_ptr, &width, &height, &bit_depth,
		     &color_type, &interlace_type, null, null);
	

	png_set_strip_16(png_ptr);	// 16 -> 8 bpc

	if (color_type & PNG_COLOR_MASK_ALPHA)
		png_set_strip_alpha(png_ptr);

	png_set_packing(png_ptr);	// 2,4 -> 8

	bool pl = false;
	
	png_color *pal;
	int numcol;
	if (color_type == PNG_COLOR_TYPE_PALETTE) {
		if (png_get_PLTE(png_ptr, info_ptr, (&pal), (&numcol)))
			pl = true;
		else
			png_set_expand(png_ptr);
	}
	set_buffers(this, pl ? dtPalette : dtRGB, width, height, 50);
	if (pl)
		memcpy(palette, pal, numcol * 3);
	png_read_update_info(png_ptr, info_ptr);

	read_file();
	free();
}
// ---------------------------------------------------------------------------

#endif // HAVE_LIBPNG