// Copyleft (Cl) 1999 by Malekith / The Final Crash of Corp. System
// <malekith@topnet.pl> / <http://malekith.topnet.pl>
// This is part of gallery image viewer. See file copying-left for details.
// Created on Aug 10 1999
// $Id: quantize.cc,v 2.15 1999/12/11 21:54:13 malekith Exp malekith $

#include "main.h"
#include "image.h"
#include "browser.h"

extern "C" void finish_doing();
extern "C" int progress_bar(int,int);
extern "C"
int rgb_to_pal(	byte * rgb_data, int w, int h, byte * indexed_data, 
		byte * rm, byte * gm, byte * bm, int descols, int dither);
// ---------------------------------------------------------------------------
void Browser::set_buffers(Image *it, int data_type, int ww, int hh, int perc)
{
	w = ww; h = hh;
	it->palette = current_palette;
	switch (data_type) {
	case dtGray:
	case dtPalette: 
		bpp = 1; 
		break;
	case dtRGB: 
		bpp = 3; 
		break;
	case dtRGBA: 
	case dtRGBA_ud: 
		bpp = 4; 
		break;
	default:
		abort();
		break;
	}
	type = data_type;
	it->line_width = bpp * w;
	it->data_size = it->line_width * h;
	data = it->data = new byte[it->data_size];
	
	if (type != dtPalette && bitspp == 8)
		now_doing(0,perc);
		
	if (w <= screen_w && h <= screen_h) {
		fs_w = w; fs_h = h;
	} else {
		fs_w = screen_w;
		if (h * fs_w / w > screen_h)	// have to recount
			fs_w = w * screen_h / h;
		fs_h = h * fs_w / w;

		fs_w <?= screen_w;	// 2 b
		fs_h <?= screen_h;	// sure ...
	}
}
// ---------------------------------------------------------------------------
void Browser::strip_alpha()
{
	if (type != dtRGBA && type != dtRGBA_ud)
		return;
	int n = w * h;
	byte *nd = new byte[n * 3];
	byte *d = nd;
	if (type == dtRGBA) {
		byte *s = data;
		while (n-- > 0) {
			*d++ = *s++;
			*d++ = *s++;
			*d++ = *s++;
			s++;
		}
	} else {
		int lw = w * 4;
		n = h;
		while (n-->0) {
			byte *s = data + lw * n;
			int m = w;
			while (m-->0) {
				*d++ = *s++;
				*d++ = *s++;
				*d++ = *s++;
				s++;
			}
		}
	}
	
	delete [] data;
	data = nd;
	type = dtRGB;
}
// ---------------------------------------------------------------------------
void Browser::do_quantize()
{
	finish_doing();
	orig.alloc(w,h);
	if (quantize == q332) {
		// genrate 332 palette
		byte *p = current_palette;
		for (int i = 0; i < 256; i++) {
			*p++ = (i & 0xe0) >> 2;
			*p++ = ((i & 0x1c) << 3) >> 2;
			*p++ = ((i & 3) << 6) >> 2;
		}
		// map data
		int m = w * h;
		int n = m;
		register byte *s = data;
		register byte *d = orig.data();
		while (n-- > 0) {
			register int k = *s++ & 0xe0;
			k += (*s++ >> 3) & 0x1c;
			k += (*s++ >> 6) & 0x03;
			*d++ = k;
			if (((int)s & 0xfff) == 0)
				if (progress_bar(n-m, m))
					break;
		}
	} else { // IJG quantize
		byte r[256], g[256], b[256];
		rgb_to_pal(data,
			   w, h, orig.data(), r, g, b, 256, 
			   dither == dNone);
		byte *p = current_palette;
		for (int i = 0; i < 256; i++) {
			*p++ = r[i] >> 2;
			*p++ = g[i] >> 2;
			*p++ = b[i] >> 2;
		}
	}
	delete [] data;
}
// ---------------------------------------------------------------------------
void Browser::buffer_stuff()
{
	strip_alpha();
	if (type == dtGray) {
		byte *p = current_palette;
		for (int i=0; i<256; i++) 
			*p++ = i, *p++ = i, *p++ = i;
		type = dtPalette;
	}
	
	switch (bitspp) {
	case 8:
		bs_8();
		break;
	case 15:
		bs_15();
		break;
	case 16:
		bs_16();
		break;
	case 24:
		bs_24();
		break;
	case 32:
		bs_32();
		break;
	default:
		abort();
	}
	
	if (fit_screen && (fs_w != w || fs_h != h))
		on_screen.resize_from(orig,fs_w,fs_h);
	else
		on_screen.copy_from(orig);
}
// ---------------------------------------------------------------------------
void Browser::bs_8()
{
	if (type == dtRGB)
		do_quantize();
	else {
		byte *p = current_palette;
		int n = 256*3;
		while (n-->0)
			*p++ >>= 2;
		orig.set_buffer(data,w,h);
	}
}
// ---------------------------------------------------------------------------
void Browser::bs_15()
{
	orig.alloc(w, h);
	int n = w * h;
	unsigned short *z = (unsigned short *) orig.data();
	register byte *p = data;
	register unsigned short a;

	if (type == dtPalette) {
		register byte *x;
		while (n-- > 0) {
			x = current_palette + (int)(*p) + (int)*p + (int)*p;
			p++;
			a = (*x++ & 0xf8) << (5 + 5 - 3);
			a |= (*x++ & 0xf8) << (5 - 3);
			a |= *x++ >> 0 + 3;
			*z++ = a;
		}
	} else {
		while (n-- > 0) {
			a = (*p++ & 0xf8) << (5 + 5 - 3);
			a |= (*p++ & 0xf8) << (5 - 3);
			a |= *p++ >> 0 + 3;
			*z++ = a;
		}
	}
}
// ---------------------------------------------------------------------------
void Browser::bs_16()
{
	orig.alloc(w, h);
	int n = w * h;
	unsigned short *z = (unsigned short *) orig.data();
	register byte *p = data;
	register unsigned short a;

	if (type == dtPalette) {
		register byte *x;
		while (n-- > 0) {
			x = current_palette + (int)(*p) + (int)*p + (int)*p;
			p++;
			a = (*x++ & 0xf8) << (5 + 6 - 3);
			a |= (*x++ & 0xfc) << (6 - 3);
			a |= *x++ >> 0 + 3;
			*z++ = a;
		}
	} else {
		while (n-- > 0) {
			a = (*p++ & 0xf8) << (5 + 6 - 3);
			a |= (*p++ & 0xfc) << (6 - 3);
			a |= *p++ >> 0 + 3;
			*z++ = a;
		}
	}
}
// ---------------------------------------------------------------------------
void Browser::bs_24()
{
	orig.alloc(w, h);
	int n = w * h;
	register byte *s = data;
	register byte *d = orig.data();

	if (type == dtPalette) {
		register byte *x;
		while (n-- > 0) {
			x = current_palette + *s + *s + *s + 2;
			s++;
			*d++ = *x--;
			*d++ = *x--;
			*d++ = *x;
		}
	} else {
		while (n-- > 0) {
			*d++ = s[2];
			*d++ = s[1];
			*d++ = s[0];			
			s += 3;
		}
	}
}
// ---------------------------------------------------------------------------
void Browser::bs_32()
{
	orig.alloc(w, h);
	int n = w * h;
	register byte *s = data;
	register byte *d = orig.data();

	if (type == dtPalette) {
		register byte *x;
		while (n-- > 0) {
			x = current_palette + *s + *s + *s + 2;
			s++;
			*d++ = *x--;
			*d++ = *x--;
			*d++ = *x;
			d++;
		}
	} else {
		while (n-- > 0) {
			*d++ = s[2];
			*d++ = s[1];
			*d++ = s[0];			
			d++;
			s += 3;
		}
	}
}
// ---------------------------------------------------------------------------
