// Copyleft (Cl) 1999 by Malekith / The Final Crash of Corp. System
// <malekith@topnet.pl> / <http://malekith.topnet.pl>
// This is part of gallery image viewer. See file copying-left for details.
// Created on Aug 10 1999
// $Id: selector.cc,v 2.3 1999/12/11 21:54:13 malekith Exp malekith $

#include "main.h"
#include "browser.h"

#include "gif.h"
#include "xpm.h"
#include "pcx.h"
#include "bmp.h"
#include "pnm.h"
#include "xbm.h"
#include "targa.h"
#include "iff.h"

#include <fnmatch.h>
#include <unistd.h>
#include <fcntl.h>

#ifdef HAVE_LIBJPEG
#	include "jpeg.h"
static JpegImage jpeg;
#endif

#ifdef HAVE_LIBPNG
#	include "png.h"
static PngImage png;
#endif

#ifdef HAVE_LIBTIFF
#	include "tiff.h"
static TiffImage tiff;
#endif

#include <ctype.h>


static GifImage gif;
static XpmImage xpm;
static PcxImage pcx;
static BmpImage bmp;
static PnmImage pnm;
static XbmImage xbm;
static TargaImage targa;
static IffImage iff;

Image * fmts[] = {
#ifdef HAVE_LIBJPEG
	&jpeg,
#endif
#ifdef HAVE_LIBPNG
	&png,
#endif
#ifdef HAVE_LIBTIFF
	&tiff,
#endif
	&gif,
	&xpm,
	&pcx,
	&bmp,
	&pnm,
	&xbm,
	&targa,
	&iff,
	null
};
int my_mkstemp(char * temp)
{
	char t[strlen(temp)+1];
	for (int i=0; i<999999; i++) {
		char * p = temp, * d = t;
		int n = i;
		while (*p) {
			if (n != 0 && *p == '#') {
				*d++ = n%10 + '0';
				n/=10;
			} else
				*d++ = *p;
			p++;
		}
		*d = 0;
		if (access(t, F_OK) != 0) {
			int fd = creat(t, 0600);
			if (fd == -1)
				throw xRecover(_("problems with create()"));
			strcpy(temp, t);
			return fd;
		}
	}
	throw xRecover("you have at least one milion files in /tmp. "
			"are you crazy or something ?");
}
char * pipe_zip(const char * zip, const char * fname)
{
	byte buffer[2048];
	char * cmd;
	asprintf(&cmd, "%s -d -c %s",zip,fname);
	if (cmd == null)
		throw xRecover("out of mem for asprintf");
	FILE * p = popen(cmd,"r");
	if (p == null)
		throw xRecover(_("can't open pipe"));
	char * tn;
	asprintf(&tn, "/tmp/######%s",fname);
	if (tn == null)
		throw xRecover("out of mem for asprintf");
		
	char * x = tn + strlen(tn);
	while (*--x != '.') *x = 0;
	*x = 0;
	
	int fd = my_mkstemp(tn);
	
	int n;
	while ((n = fread(buffer, 1, sizeof(buffer), p)) != 0 && n != EOF)
		write(fd, buffer, n);
	close(fd);
	pclose(p);
	free(cmd);
	return tn;
}
Image *match_by_name(const char * name)
{
	Image ** p = fmts;
	while (*p) {
		if ((*p)->match_name(name))
			return *p;
		p++;
	}
	return null;
}
Image *match_by_magic(const char * name)
{
	byte magic_buf[1024];
	FILE * f = fopen(name,"r");
	if (f == null)
		return null;
	fread(magic_buf,1,sizeof(magic_buf),f);
	fclose(f);
	Image ** p = fmts;
	while (*p) {
		if ((*p)->match_magic(magic_buf))
			return *p;
		p++;
	}
	return null;
}
void fmt_text(int maxn, char * s);
void Browser::display_formats()
{
	printf(_("Supported formats:\n"));
	Image **p = fmts;
	int n = 80;
	if (getenv("COLUMNS") && atoi(getenv("COLUMNS")) > 0)
		n = atoi(getenv("COLUMNS"));
	while (*p) {
		char *t;
		asprintf(&t, "  \e[1;34m* \e[1;36m%s \e[0m(\e[0;31m%s\e[0m) "
			     "%s \e[1;34m%s\e[0m\n",
			 (*p)->get_fmt_name(), (*p)->get_fmt_cmt(), 
			 _("from"), _((*p)->get_fmt_src()));
		if (t == null)
			throw xDead("no mem for asprintf");
		fmt_text(n, t);
		printf(t);
		free(t);
		p++;
	}
}
Image *Browser::selector(const char *filename, bool force_magic)
{
	file_is_temp = false;
	const char *fn = filename;
	if (fnmatch("*.gz", fn, FNM_LEADING_DIR | FNM_CASEFOLD) == 0)
		fn = pipe_zip("gzip", fn);
	if (fnmatch("*.bz2", fn, FNM_LEADING_DIR | FNM_CASEFOLD) == 0)
		fn = pipe_zip("bzip2", fn);
	if (fnmatch("*.Z", fn, FNM_LEADING_DIR | FNM_CASEFOLD) == 0)
		fn = pipe_zip("compress", fn);
		
	Image *loader = null;
	if (!force_magic)
		loader = match_by_name(fn);
	
	if (loader == null) 
		loader = match_by_magic(fn);
	
	if (loader == null) {
		name = null;
		throw xRecover(_("file format not supported"));
	} else {
		name = new_str(fn);
		if (fn != filename) {
			file_is_temp = true;
			free((void*)fn);
		}
		return loader;
	}
	
}
