// Copyleft (Cl) 1999 by Malekith / The Final Crash of Corp. System
// <malekith@topnet.pl> / <http://malekith.topnet.pl>
// This is part of gallery image viewer. See file copying-left for details.
// $Id: sim_argp.cc,v 1.18 1999/12/11 21:54:13 malekith Exp malekith $

// simulates argp parser
// actaully it's used even if it's present...

#include <stdarg.h>
#include <ctype.h>

struct argp_option {
	const char *name;
	int key;
	const char *arg;
	int flags;		// ignored
	const char *doc;
};
typedef error_t (*argp_parser_t) (int key, char *arg,
				  struct argp_state * state);

#define ARGP_KEY_END		0x1000001
#define ARGP_KEY_INIT		0x1000003
#define ARGP_KEY_FINI		0x1000007
#define ARGP_KEY_NO_ARGS	0x1000002
#define ARGP_KEY_ARG		0
#define ARGP_ERR_UNKNOWN	0 // ignored!

struct argp {
	const struct argp_option *options;
	argp_parser_t parser;
	const char *args_doc;
	const char *doc;
};
struct argp_state {
	void * input;
	const char * pname;
};
const char *argp_program_version;
const char *argp_program_bug_address;

void argp_error (const struct argp_state * state, const char * fmt, ... )
{
	fprintf(stderr,"%s: ",state->pname);
	va_list ap;
	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	fprintf(stderr,_("\nTry %s --help for more info\n"),state->pname);
	exit(1);
}
void fmt_text(int maxn, char * s);
void argp_printa(int column, const char * text)
{
	char buf[strlen(text)+1];
	strcpy(buf, text);
	fmt_text(78-column, buf);
	printf("\r\e[%dC",column);
	char * p = buf;
	while (*p) {
		if (*p == '\n')
			printf("\n\e[%dC",column);
		else
			putchar(*p);
		p++;
	}
}
void argp_help(const struct argp *ap, struct argp_state * s)
{
	const argp_option *o = ap->options;
	printf(_("Usage: %s [OPTION...] %s\n"), s->pname, _(ap->args_doc));
	const char * p = _(ap->doc);
	while (p && *p && *p!='\v')
			putchar(*p++);
	printf("\n\n");
	const int clm = 30;
	while(o->name) {
		printf("  --%s",o->name);
		if (o->arg != null)
			printf("=%s",o->arg);
		if (isalnum(o->key))
			printf(", -%c",o->key);
		argp_printa(clm,_(o->doc));
		putchar('\n');
		o++;
	}
	printf("\n  --help, -?"); argp_printa(clm,_("This help."));
	printf("\n  --version"); argp_printa(clm,_("Displays version."));
	printf("\n\n");
	if (p && *p == '\v') p++;
	while (p && *p)
		putchar(*p++);
	printf(_("\nReport bugs to %s.\n"),argp_program_bug_address);
	exit(0);
}

void handle_long(const struct argp *ap, const char *pp, struct argp_state * s)
{
	char buf[strlen(pp)+1];
	char * p = buf;
	strcpy(p,pp);
	
	const argp_option *o = ap->options;
	if (strcmp(p,"help") == 0)
		argp_help(ap,s);
	if (strcmp(p,"version") == 0) {
		printf("%s\n",argp_program_version);
		exit(0);
	}
	while (o->name != null) {
		if (strncmp(p, o->name, strlen(o->name)) == 0)
			break;
		o++;
	}
	if (o->name) {
		p += strlen(o->name);
		if (o->arg != null && (*p != '=' || *++p == '\0'))
			argp_error(s,_("option `--%s' requires an argument"),
				   o->name);
		if (o->arg == null && *p != '\0')
			argp_error(s,_("option `--%s' doesn't allow an argument"),
				   o->name);
		ap->parser(o->key, *p ? p : null, s);
	} else
		argp_error(s,_("unreconized option `--%s'"),p);
}
void handle_short(const struct argp *ap, const char *pp, struct argp_state *s,
		  char **&argv)
{
	char buf[strlen(pp)+1];
	char *p = buf;
	strcpy(p,pp);
	
	const argp_option *o = ap->options;
	if (*p == '?')
		argp_help(ap,s);
		
	while (o->name != null) {
		if (*p == o->key)
			break;
		o++;
	}
	if (o->name) {
		p++;
		if (o->arg && *p == 0) {
			if (*++argv == null)
				argp_error(s,_("option requires an argument "
						"-- %c"), o->key);
			ap->parser(o->key, *argv, s);
		} else if (o->arg)
			ap->parser(o->key, p, s);
		else {
			ap->parser(o->key, null, s);
			if (*p)
				handle_short(ap, p, s, argv);
		}
	} else
		argp_error(s, _("invalid option -- %c"), *p);
}

error_t argp_parse(const struct argp *ap,
		   int /*argc*/, char **argv,
		   unsigned /*flags */ , int /*arg_index */ ,
		   void *input)
{
	argp_state s = { input, argv[0] };
	char ** p = argv + 1;
	bool no_args = true;
	ap->parser(ARGP_KEY_INIT, null, &s);
	while (*p) {
		if ((*p)[0] == '-') {
			if ((*p)[1]=='-')
				handle_long(ap, *p+2, &s);
			else
				handle_short(ap, *p+1, &s, p);
		} else {
			ap->parser(ARGP_KEY_ARG, *p, &s);
			no_args = false;
		}
		p++;
	}
	if (no_args)
		ap->parser(ARGP_KEY_NO_ARGS, null, &s);
	ap->parser(ARGP_KEY_FINI, null, &s);
	return 0;
}
