/*         ______   ___    ___
 *        /\  _  \ /\_ \  /\_ \
 *        \ \ \L\ \\//\ \ \//\ \      __     __   _ __   ___
 *         \ \  __ \ \ \ \  \ \ \   /'__`\ /'_ `\/\`'__\/ __`\
 *          \ \ \/\ \ \_\ \_ \_\ \_/\  __//\ \L\ \ \ \//\ \L\ \
 *           \ \_\ \_\/\____\/\____\ \____\ \____ \ \_\\ \____/
 *            \/_/\/_/\/____/\/____/\/____/\/___L\ \/_/ \/___/
 *                                           /\____/
 *                                           \_/__/
 *      By Shawn Hargreaves,
 *      1 Salisbury Road,
 *      Market Drayton,
 *      Shropshire,
 *      England, TF9 1AJ.
 *
 *	WWW: http://www.talula.demon.co.uk/allegro/
 *	e-mail: shawn@talula.demon.co.uk   
 *
 *      TGA reader by Tim Gunn.
 *
 *      RLE support added by Michal Mertl and Salvador Eduardo Tropea.
 *
 *	Copyright :
 *
 *  Allegro is swap-ware. You may use, modify, redistribute, and generally 
 *  hack it about in any way you like, but if you do you must send me 
 *  something in exchange. This could be a complimentary copy of a game, an 
 *  addition or improvement to Allegro, a bug report, some money (this is 
 *  particularly encouraged if you use Allegro in a commercial product), or 
 *  just a copy of your autoexec.bat if you don't have anything better. If 
 *  you redistribute parts of Allegro or make a game using it, it would be 
 *  nice if you mentioned me somewhere in the credits, but if you just want 
 *  to pinch a few routines that is OK too. I'll trust you not to rip me off.
 *
 */

/*
 * modified by malekith
 * telling the truth, only algorithm remains the same ...
 * $Id: tga.c,v 1.15 1999/09/25 20:53:09 malekith Exp malekith $
 */

#include "main-c.h"

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>

#include "tga.h"

typedef unsigned char byte;
const int true = 1;
const int false = 0;

int progress_bar(int,int);

int igetw(FILE * f)
{
	int b1, b2;
	if ((b1 = getc(f)) != EOF && (b2 = getc(f)) != EOF)
		return ((b2 << 8) | b1);
	return EOF;
}
int igetl(FILE *f) {
   int b1, b2, b3, b4;

   if ((b1 = getc(f)) != EOF && (b2 = getc(f)) != EOF &&
       (b3 = getc(f)) != EOF && (b4 = getc(f)) != EOF )
	       return (((long)b4 << 24) | ((long)b3 << 16) |
		       ((long)b2 << 8) | (long)b1);
   return EOF;
}
// calls progress_bar if needed
#define prog() ({ if (w > 10000 && (c & 0x1ff) == 0) progress_bar(c,w); })
/* rle_tga_read:
 *  Helper for reading 256 color RLE data from TGA files.
 */
static void rle_tga_read(byte *b, int w, FILE * f)
{
	byte value;
	int count;
	int c = 0;

	do {
		count = getc(f);
		if (count & 0x80) {
			count = (count & 0x7F) + 1;
			c += count;
			value = getc(f);
			while (count--)
				*(b++) = value;
		} else {
			count++;
			c += count;
			fread(b, 1, count, f);
			b += count;
		}
		prog();
	} while (c < w);
}

/* rle_tga_read24:
 *  Helper for reading 24 bit RLE data from TGA files.
 */
static void rle_tga_read24(byte *b, int w, FILE * f)
{
	byte value[4];
	int count;
	int c = 0;

	do {
		count = getc(f);
		if (count & 0x80) {
			count = (count & 0x7F) + 1;
			c += count;
			fread(value, 1, 3, f);
			while (count--) {
				*b++ = value[2];
				*b++ = value[1];
				*b++ = value[0];
			}
		}

		else {
			count++;
			c += count;
			while (count--) {
				fread(value, 1, 3, f);
				*b++ = value[2];
				*b++ = value[1];
				*b++ = value[0];
			}
		}
		prog();
	} while (c < w);
}

static void tga_read24(byte *b, int w, FILE * f)
{
	byte value[4];
	while (w-->0) {
		if (fread(value, 1, 3, f) <= 0)
			break;
		*b++ = value[2];
		*b++ = value[1];
		*b++ = value[0];
	} 
}


static void tga_read16(byte *b, int w, FILE * f)
{
	while (w-->0) {
		unsigned int k = igetw(f);
		*b++ = (k >> 10) & 0x1f;
		*b++ = (k >> 5)  & 0x1f;
		*b++ = (k >> 0)  & 0x1f;
	}
}
/* rle_tga_read16:
 *  Helper for reading 16 bit RLE data from TGA files.
 */
static void rle_tga_read16(byte * b, int w, FILE * f)
{
	unsigned int value;
	int count;
	int c = 0;

	do {
		count = getc(f);
		if (count & 0x80) {
			count = (count & 0x7F) + 1;
			c += count;
			value = igetw(f);

			while (count-- > 0) {
				*b++ = (value >> 10) & 0x1f;
				*b++ = (value >> 5) & 0x1f;
				*b++ = (value >> 0) & 0x1f;
			}
		}

		else {
			count++;
			c += count;
			while (count-- > 0) {
				value = igetw(f);
				*b++ = (value >> 10) & 0x1f;
				*b++ = (value >> 5) & 0x1f;
				*b++ = (value >> 0) & 0x1f;
			}
		}
	} while (c < w);
}
struct tga_header {
	byte   id_length;
	byte   palette_type;
	byte   image_type;
	unsigned short first_color;
	unsigned short palette_colors;
	byte   palette_entry_size;
	unsigned short left;
	unsigned short top;
	unsigned short width;
	unsigned short height;
	byte   bpp;
	byte   descriptor_bits;
} __attribute__ ((packed));

int tga_get_header(const char *fname, image_info * info)
{
	struct tga_header head;
	FILE *f = fopen(fname, "r");
	int n;
	if (!f)
		return false;
	n = fread(&head, 1, sizeof(head), f);
	fclose(f);
	if ( n != sizeof(head) || !( 
		head.bpp == 8 ||
		head.bpp == 15 ||
		head.bpp == 16 ||
		head.bpp == 24) ){
		return false;
	}
	info->width = head.width;
	info->height = head.height;
	info->bpp = head.bpp;
	return true;
}

/* load_tga:
 *  Loads a 256 color or 24 bit uncompressed TGA file, returning a bitmap
 *  structure and storing the pallete data in the specified pallete (this
 *  should be an array of at least 256 RGB structures).
 */
int tga_load(const char *filename, int blen, byte * dbuf, byte * palbuf)
{
	byte image_id[256], image_palette[256][3];
	int compressed, ok;
	int y,i;
	struct tga_header head;
	FILE *f;
	
	f = fopen(filename, "r");
	if (!f)
		return false;
	fread(&head,1,sizeof(head),f);
	fread(image_id, 1, head.id_length, f);
	fread(image_palette, 3, head.palette_colors, f);
	if (feof(f)) {
		fclose(f);
		return false;
	}
	/* Image type:
	 *    0 = no image data
	 *    1 = uncompressed color mapped
	 *    2 = uncompressed true color
	 *    3 = grayscale
	 *    9 = RLE color mapped
	 *   10 = RLE true color
	 *   11 = RLE grayscale
	 */
	compressed = (head.image_type & 8) != 0;
	head.image_type &= 7;
	if ((head.image_type < 1) || (head.image_type > 3) || 
		(blen < head.width * head.height)) {
		fclose(f);
		return false;
	}
	switch (head.image_type) {
	case 1:
		/* paletted image */
		if ((head.palette_type != 1) || (head.bpp != 8)) {
			fclose(f);
			return false;
		}
		for (i = 0; i < head.palette_colors; i++) {
			*palbuf++ = image_palette[i][2];
			*palbuf++ = image_palette[i][1];
			*palbuf++ = image_palette[i][0];
		}
		break;
	case 2:
		/* truecolor image */
		if (head.bpp == 16)
			head.bpp = 15;
			
		if (head.palette_type != 0 || 
		    (head.bpp != 24 && head.bpp != 15) ||
		    blen < head.width * head.height * 3){
			fclose(f);
			return false;
		}
		break;
	case 3:
		/* grayscale image */
		if ((head.palette_type != 0) || (head.bpp != 8)) {
			fclose(f);
			return false;
		}
		for (i = 0; i < 256; i++) {
			*palbuf++ = (byte) i;
			*palbuf++ = (byte) i;
			*palbuf++ = (byte) i;
		}
		break;
	default:
		fclose(f);
		return false;
	}
	if ((head.descriptor_bits & 0x20) && compressed) {
		if (head.bpp == 24)
			rle_tga_read24(dbuf, head.width*head.height, f);
		else if (head.bpp == 15)
			rle_tga_read16(dbuf, head.width*head.height, f);
		else 
			rle_tga_read(dbuf, head.width*head.height, f);
	} else for (y = head.height; y; y--) {
		int yc = (head.descriptor_bits & 0x20) ? head.height - y : y - 1;
		byte * p = dbuf + yc * head.width * (head.bpp == 8 ? 1 : 3);
		if (compressed) {
			if (head.bpp == 24)
				rle_tga_read24(p, head.width, f);
			else if (head.bpp == 15)
				rle_tga_read16(p, head.width, f);
			else 
				rle_tga_read(p, head.width, f);
		} else {
			if (head.bpp == 24)
				tga_read24(p, head.width, f);
			else if (head.bpp == 15)
				tga_read16(p, head.width, f);
			else 
				fread(p, 1, head.width, f);
		}
		if (progress_bar(head.height-y,head.height))
			break;
	}
	ok = !ferror(f);
	fclose(f);
	return ok;
}
